/*
 * Apache auth_ldap_log_reason format string vulnerabilty
 *
 * Copyright (C) 2007 Sourcefire, Inc. All Rights Reserved
 *
 * Writen by Patrick Mullen, Sourcefire VRT <pmullen@sourcefire.com>
 *
 * This file may contain proprietary rules that were created, tested and
 * certified by Sourcefire, Inc. (the "VRT Certified Rules") as well as
 * rules that were created by Sourcefire and other third parties and
 * distributed under the GNU General Public License (the "GPL Rules").  The
 * VRT Certified Rules contained in this file are the property of
 * Sourcefire, Inc. Copyright 2005 Sourcefire, Inc. All Rights Reserved.
 * The GPL Rules created by Sourcefire, Inc. are the property of
 * Sourcefire, Inc. Copyright 2002-2005 Sourcefire, Inc. All Rights
 * Reserved.  All other GPL Rules are owned and copyrighted by their
 * respective owners (please see www.snort.org/contributors for a list of
 * owners and their respective copyrights).  In order to determine what
 * rules are VRT Certified Rules or GPL Rules, please refer to the VRT
 * Certified Rules License Agreement.
 */

#include "sf_snort_plugin_api.h"
#include "sf_snort_packet.h"
#include "web-misc_base64_decode.h"

//#define DEBUG
#ifdef DEBUG
#define DEBUG_WRAP(code) code
#else
#define DEBUG_WRAP(code)
#endif


// backfill support for older versions -- http client headers and getBuffer()
#ifndef CONTENT_BUF_HEADER
#define CONTENT_BUF_HEADER CONTENT_BUF_RAW
#endif

#ifndef CURSOR_IN_BOUNDS
#define CURSOR_IN_BOUNDS 1
#endif

#ifdef MISSING_GETBUFFER
#ifndef EMULATE_GETBUFFER
#define EMULATE_GETBUFFER
int getBuffer(void *packet, int flags, const u_int8_t **start, const u_int8_t **end) {
   SFSnortPacket *sp = (SFSnortPacket *)packet;

   DEBUG_WRAP(printf("using emulated getBuffer()\n"));

   // Always returns pointers based upon CONTENT_BUF_RAW, ignoring flags
   *start = sp->payload;
   *end = sp->payload + sp->payload_size;

   return CURSOR_IN_BOUNDS;
}
#endif
#endif

/* declare detection functions */
int ruleAPACHEAUTHLDAPeval(void *p);

/* declare rule data structures */
/* precompile the stuff that needs pre-compiled */
/* flow:established,to_client; */
static FlowFlags ruleAPACHEAUTHLDAPflow0 = 
{
    FLOW_ESTABLISHED | FLOW_TO_SERVER
};

static RuleOption ruleAPACHEAUTHLDAPoption0 =
{
    OPTION_TYPE_FLOWFLAGS,
    {
        &ruleAPACHEAUTHLDAPflow0
    }
};

// content:"Authorization:"; 
static ContentInfo ruleAPACHEAUTHLDAPcontent0 = 
{
    (u_int8_t*)"Authorization:", /* pattern (now in snort content format) */
    0, /* depth */
    0, /* offset */
    CONTENT_BUF_HEADER | CONTENT_NOCASE, /* flags */ // XXX - need to add CONTENT_FAST_PATTERN support
    NULL, /* holder for boyer/moore PTR */
    NULL, /* more holder info - byteform */
    0 /* byteform length */
};

static RuleOption ruleAPACHEAUTHLDAPoption1 = 
{
    OPTION_TYPE_CONTENT,
    {
        &ruleAPACHEAUTHLDAPcontent0
    }
};

// pcre:"/^Authorization:\s*Basic/smi"; 
static PCREInfo ruleAPACHEAUTHLDAPpcre0 = 
{
    "^Authorization:\\s*Basic\\s+", /* pattern (now in snort content format) */
    0, /* compiled expression */
    0, /* compiled extra */
    PCRE_CASELESS | PCRE_MULTILINE | PCRE_DOTALL, /* compile flags */
    CONTENT_BUF_HEADER /* flags */ // XXX - need to add CONTENT_FAST_PATTERN support
};

static RuleOption ruleAPACHEAUTHLDAPoption2 = 
{
    OPTION_TYPE_PCRE,
    {
        &ruleAPACHEAUTHLDAPpcre0
    }
};

#if 0
// Removed pcre_exec because our check is so simple

// pcre:"/%[0-9]*\.?[0-9]*[:formatspecifiers:]/";
static PCREInfo ruleAPACHEAUTHLDAPpcre1 =
{
//    "%[-# +'I]*[0-9]*\\.?[0-9]*[qjzthdiouxefgcrslnp]", /* pattern (now in snort content format) */ // ZDNOTE 
//    "%.+%.", 	/* regex.  The above is technically more correct, but this is faster and good enough */
    "^[^:]*%",    /* regex.  The new, even looser version. Finds any % in the username (before : separator)*/
    0, /* compiled expression */
    0, /* compiled extra */
    0, /* compile flags */
    CONTENT_BUF_HEADER /* flags */ // XXX - need to add CONTENT_FAST_PATTERN support
};

static RuleOption ruleAPACHEAUTHLDAPoption3 =
{
    OPTION_TYPE_PCRE,
    {
        &ruleAPACHEAUTHLDAPpcre1
    }
};
#endif

/* references for.ruleid APACHEAUTHLDAP */
static RuleReference ruleAPACHEAUTHLDAPref0 =
{
    "cve", /* type */
    "2006-0150" /* value */
};

static RuleReference ruleAPACHEAUTHLDAPref1 =
{
    "bugtraq", /* type */
    "16177" /* value */
};

static RuleReference *ruleAPACHEAUTHLDAPrefs[] =
{
    &ruleAPACHEAUTHLDAPref0,
    &ruleAPACHEAUTHLDAPref1,
    NULL
};

RuleOption *ruleAPACHEAUTHLDAPoptions[] =
{
    &ruleAPACHEAUTHLDAPoption0,
    &ruleAPACHEAUTHLDAPoption1,
    &ruleAPACHEAUTHLDAPoption2,
// removed pcre_exec    &ruleAPACHEAUTHLDAPoption3,
    NULL
};


Rule ruleAPACHEAUTHLDAP = {
   /* rule header, akin to => tcp any any -> any any */
   {
       IPPROTO_TCP, /* proto */
       EXTERNAL_NET, /* SRCIP     */
       "any", /* SRCPORT   */
       0, /* DIRECTION */
       HTTP_SERVERS, /* DSTIP     */
       HTTP_PORTS, /* DSTPORT   */
   },
   /* metadata */
   { 
       3,  /* genid (HARDCODED!!!) */
       13308, /* sid fbdc7055-77a0-4ea7-9bcc-933448214c84 */
       5, /* revision f6c43a4a-8a31-48e3-a942-42070a209a19 */
       "attempted-user", /* classification */
       0,  /* hardcoded priority XXX NOT PROVIDED BY GRAMMAR YET! */
       "WEB-MISC Apache auth_ldap_log_reason format string attempt",     /* message */
       ruleAPACHEAUTHLDAPrefs /* ptr to references */
#ifdef HAS_METADATA
        ,NULL
#endif
   },
   ruleAPACHEAUTHLDAPoptions, /* ptr to rule options */
   &ruleAPACHEAUTHLDAPeval, /* use the built in detection function */
   0, /* am I initialized yet? */
   0, /* number of options */
   0,  /* don't alert */
   NULL /* ptr to internal data... setup during rule registration */
};

/* detection functions */
int ruleAPACHEAUTHLDAPeval(void *p) {
   const u_int8_t *cursor = 0;
   const u_int8_t *end_of_buffer;

   SFSnortPacket *sp = (SFSnortPacket *) p;

   // Base64 stuff
   u_int8_t base64buf[256], decodedbuf[256], *decodedbuf_ptr;
   u_int32_t inputchars, base64bytes, decodedbytes;

/* Removing pcre_exec since the check is so simple now
   // manual pcre stuff
   int result;
   int ovector[3];  // Needs to be a multiple of 3
*/

   if(sp == NULL)
      return RULE_NOMATCH;

   if(sp->payload == NULL)
      return RULE_NOMATCH;

   DEBUG_WRAP(printf("ruleAPACHEAUTHLDAPeval start\n"));

   if(getBuffer(sp, CONTENT_BUF_HEADER, &cursor, &end_of_buffer) != CURSOR_IN_BOUNDS)
      return RULE_NOMATCH;

   // flow:established,to_server;
   if(checkFlow(p, ruleAPACHEAUTHLDAPoptions[0]->option_u.flowFlags) <= 0 )
      return RULE_NOMATCH;

   DEBUG_WRAP(printf("passed flow check\n"));

   // content:"Authorization:";
   if(contentMatch(p, ruleAPACHEAUTHLDAPoptions[1]->option_u.content, &cursor) <= 0)
      return RULE_NOMATCH;

   DEBUG_WRAP(printf("found content:\"Authorization:\" %p\n", cursor));

   // pcre:"/^Authorization:\s*Basic\s+/mi"
   if(pcreMatch(p, ruleAPACHEAUTHLDAPoptions[2]->option_u.pcre, &cursor) <= 0)
      return RULE_NOMATCH;

   DEBUG_WRAP(printf("found pcre:\"/^Authorization:\\s*Basic\\s+/mi\" %p\n", cursor));

   // At this point, cursor should point to the start of the auth data
   inputchars = (end_of_buffer > cursor + sizeof(base64buf)) ? sizeof(base64buf) : end_of_buffer - cursor; 
   if(unfold_header(cursor, inputchars, base64buf, sizeof(base64buf), &base64bytes) != 0)
      return RULE_NOMATCH;
  
   DEBUG_WRAP(printf("Successfully unfolded header (%s)(%d)\n", base64buf, base64bytes));

   if(base64decode(base64buf, base64bytes, decodedbuf, sizeof(decodedbuf), &decodedbytes) < 0)
      return RULE_NOMATCH;

   DEBUG_WRAP(printf("Successfully base64 decoded (%s)(%d)\n", decodedbuf, decodedbytes));

   // Now run our regex on the base64 decoding to find an attack
/* Doing this by hand now that our pcre is super simple
   result = pcre_exec(ruleAPACHEAUTHLDAPoptions[3]->option_u.pcre->compiled_expr,  // pcre_compile()
                      ruleAPACHEAUTHLDAPoptions[3]->option_u.pcre->compiled_extra, // pcre_study()
                      (char*)decodedbuf,        // subject string
                      decodedbytes,      // subject length
                      0,                 // start offset
                      0,                 // options (handled at compile time)
                      ovector,           // ovector for storing result substrings
                      sizeof(ovector));  // size of ovector

   DEBUG_WRAP(printf("result = %d\n", result));

   if(result >= 0)
      return RULE_MATCH;
*/

   // Manually go through the string, looking for a '%' before the ':' separator.
   decodedbuf_ptr = decodedbuf;
   // Bug 47146 -- removing check for uname:passwd separator to detect format
   // string in passwords.  Technically, this is not the vuln, but we're doing
   // it anyway for certification.  This will false positive if someone uses a
   // percent character in their password for HTTP Basic Auth
   while(*decodedbuf_ptr != '\0' /*&& *decodedbuf_ptr != ':'*/) {
      if(*decodedbuf_ptr++ == '%') 
         return RULE_MATCH;
   }
       
   return RULE_NOMATCH;
}

